from __future__ import annotations

import logging
import os
from typing import Optional

from soda_core.common.sql_ast import DROP_TABLE, DROP_VIEW
from soda_core.common.statements.metadata_tables_query import (
    FullyQualifiedTableName,
    MetadataTablesQuery,
)
from soda_core.common.statements.table_types import FullyQualifiedViewName, TableType
from soda_sqlserver.test_helpers.sqlserver_data_source_test_helper import (
    SqlServerDataSourceTestHelper,
)

logger = logging.getLogger(__name__)


class FabricDataSourceTestHelper(SqlServerDataSourceTestHelper):
    def _create_database_name(self) -> Optional[str]:
        return os.getenv("FABRIC_DATABASE", "soda-ci-fabric-warehouse")

    def _create_data_source_yaml_str(self) -> str:
        """
        Called in _create_data_source_impl to initialized self.data_source_impl
        self.database_name and self.schema_name are available if appropriate for the data source type
        """
        return f"""
            type: fabric
            name: {self.name}
            connection:
                host: '{os.getenv("FABRIC_HOST", "localhost")}'
                port: '{os.getenv("FABRIC_PORT", "1433")}'
                database: '{os.getenv("FABRIC_DATABASE", "soda-ci-fabric-warehouse")}'
                authentication: '{os.getenv("FABRIC_AUTHENTICATION_TYPE", "activedirectoryserviceprincipal")}'
                client_id: '{os.getenv("FABRIC_CLIENT_ID")}'
                client_secret: '{os.getenv("FABRIC_CLIENT_SECRET")}'
                trust_server_certificate: true
                driver: '{os.getenv("FABRIC_DRIVER", "ODBC Driver 18 for SQL Server")}'
        """

    def drop_schema_if_exists(self, schema: str) -> None:
        """Drop all objects in the schema first, then drop the schema.

        SQL Server/Fabric does not support CASCADE on DROP SCHEMA, so all
        tables and views must be removed before the schema can be dropped.
        """
        try:
            if not self._does_schema_exist(schema):
                return

            dialect = self.data_source_impl.sql_dialect
            database_name = self.extract_database_from_prefix()

            metadata_tables_query: MetadataTablesQuery = self.data_source_impl.create_metadata_tables_query()
            fully_qualified_objects = metadata_tables_query.execute(
                database_name=database_name,
                schema_name=schema,
                types_to_return=[TableType.TABLE, TableType.VIEW],
            )

            # Drop views first
            for obj in fully_qualified_objects:
                if isinstance(obj, FullyQualifiedViewName):
                    view_identifier = f"{dialect.quote_default(obj.database_name)}.{dialect.quote_default(obj.schema_name)}.{dialect.quote_default(obj.view_name)}"
                    self.data_source_impl.execute_update(dialect.build_drop_view_sql(DROP_VIEW(view_identifier)))

            # Then drop tables
            for obj in fully_qualified_objects:
                if isinstance(obj, FullyQualifiedTableName):
                    table_identifier = f"{dialect.quote_default(obj.database_name)}.{dialect.quote_default(obj.schema_name)}.{dialect.quote_default(obj.table_name)}"
                    self.data_source_impl.execute_update(dialect.build_drop_table_sql(DROP_TABLE(table_identifier)))

            # Finally drop the schema
            self.data_source_impl.execute_update(f"DROP SCHEMA {dialect.quote_default(schema)};")
        except Exception as e:
            logger.warning(f"Error dropping test schema: {e}")



from __future__ import annotations

import logging
import struct
from abc import ABC
from datetime import datetime, timezone
from typing import Literal, Union

import pyodbc
from pydantic import Field
from soda_core.common.logging_constants import soda_logger
from soda_core.model.data_source.data_source import DataSourceBase
from soda_sqlserver.common.data_sources.sqlserver_data_source_connection import (
    SqlServerActiveDirectoryAuthentication,
    SqlServerActiveDirectoryInteractiveAuthentication,
    SqlServerActiveDirectoryPasswordAuthentication,
    SqlServerActiveDirectoryServicePrincipalAuthentication,
    SqlServerConnectionProperties,
    SqlServerDataSourceConnection,
    SqlServerPasswordAuth,
)

logger: logging.Logger = soda_logger


CONTEXT_AUTHENTICATION_DESCRIPTION = "Use context authentication"


# All of these classes are just copies of the SQLServerConnectionProperties classes, but with the Fabric type
class FabricConnectionProperties(SqlServerConnectionProperties, ABC):
    pass


class FabricPasswordAuth(SqlServerPasswordAuth, FabricConnectionProperties):
    pass


class FabricActiveDirectoryAuthentication(SqlServerActiveDirectoryAuthentication, FabricConnectionProperties):
    pass


class FabricActiveDirectoryInteractiveAuthentication(
    SqlServerActiveDirectoryInteractiveAuthentication, FabricConnectionProperties
):
    pass


class FabricActiveDirectoryPasswordAuthentication(
    SqlServerActiveDirectoryPasswordAuthentication, FabricConnectionProperties
):
    pass


class FabricActiveDirectoryServicePrincipalAuthentication(
    SqlServerActiveDirectoryServicePrincipalAuthentication, FabricConnectionProperties
):
    pass


class FabricDataSource(DataSourceBase, ABC):
    type: Literal["fabric"] = Field("fabric")

    connection_properties: Union[
        FabricPasswordAuth,
        FabricActiveDirectoryInteractiveAuthentication,
        FabricActiveDirectoryPasswordAuthentication,
        FabricActiveDirectoryServicePrincipalAuthentication,
    ] = Field(..., alias="connection", description="Fabric connection configuration")


def handle_datetime2(dto_value):
    tup = struct.unpack("<6hI", dto_value)  # e.g., (2017, 3, 16, 10, 35, 18, 500000000)
    return datetime(
        tup[0],
        tup[1],
        tup[2],
        tup[3],
        tup[4],
        tup[5],
        tup[6] // 1000,
        timezone.utc,  # Timezone is not supported by Fabric, so we assume that all timestamps are in UTC (i.e. we do no conversion!)
    )


class FabricDataSourceConnection(SqlServerDataSourceConnection):
    def _get_autocommit_setting(self) -> bool:
        return True  # Fabric requires autocommit to be True.

    def _create_connection(self, config: SqlServerConnectionProperties):
        my_connection = super()._create_connection(config)
        my_connection.add_output_converter(pyodbc.SQL_TYPE_TIMESTAMP, handle_datetime2)
        return my_connection



import logging
from datetime import datetime, timezone
from typing import Optional

from soda_core.common.data_source_connection import DataSourceConnection
from soda_core.common.logging_constants import soda_logger
from soda_core.common.metadata_types import SodaDataTypeName, SqlDataType
from soda_core.common.sql_ast import CREATE_TABLE_COLUMN, INSERT_INTO, VALUES_ROW
from soda_core.common.sql_dialect import SqlDialect
from soda_fabric.common.data_sources.fabric_data_source_connection import (
    FabricDataSource as FabricDataSourceModel,
)
from soda_fabric.common.data_sources.fabric_data_source_connection import (
    FabricDataSourceConnection,
)
from soda_sqlserver.common.data_sources.sqlserver_data_source import (
    SqlServerDataSourceImpl,
    SqlServerSqlDialect,
)

logger: logging.Logger = soda_logger


class FabricDataSourceImpl(SqlServerDataSourceImpl, model_class=FabricDataSourceModel):
    def __init__(self, data_source_model: FabricDataSourceModel, connection: Optional[DataSourceConnection] = None):
        super().__init__(data_source_model=data_source_model, connection=connection)

    def _create_sql_dialect(self) -> SqlDialect:
        return FabricSqlDialect()

    def _create_data_source_connection(self) -> DataSourceConnection:
        return FabricDataSourceConnection(
            name=self.data_source_model.name, connection_properties=self.data_source_model.connection_properties
        )


class FabricSqlDialect(SqlServerSqlDialect, sqlglot_dialect="fabric"):
    SODA_DATA_TYPE_SYNONYMS = (
        *SqlServerSqlDialect.SODA_DATA_TYPE_SYNONYMS,
        (SodaDataTypeName.TIMESTAMP_TZ, SodaDataTypeName.TIMESTAMP),
    )

    @classmethod
    def is_same_soda_data_type_with_synonyms(cls, expected: SodaDataTypeName, actual: SodaDataTypeName) -> bool:
        if expected == SodaDataTypeName.TIMESTAMP and actual == SodaDataTypeName.VARCHAR:
            logger.debug(
                f"In is_same_soda_data_type_with_synonyms, expected {expected} and actual {actual} are treated as the same because of Fabric cursor not distinguishing between varchar and datetime2"
            )
            return True
        return super().is_same_soda_data_type_with_synonyms(expected, actual)

    def sql_expr_timestamp_truncate_day(self, timestamp_literal: str) -> str:
        return f"DATETIMEFROMPARTS((datepart(YEAR, {timestamp_literal})), (datepart(MONTH, {timestamp_literal})), (datepart(DAY, {timestamp_literal})), 0, 0, 0, 0)"

    def _build_insert_into_values_sql(self, insert_into: INSERT_INTO) -> str:
        values_sql: str = "\n" + "\nUNION ALL ".join(
            [self._build_insert_into_values_row_sql(value) for value in insert_into.values]
        )
        return values_sql

    def _build_insert_into_values_row_sql(self, values: VALUES_ROW) -> str:
        values_sql: str = "SELECT " + ", ".join([self.literal(value) for value in values.values])
        return values_sql

    def default_casify(self, identifier: str) -> str:
        return identifier.upper()

    def literal_datetime_with_tz(self, datetime: datetime):
        # Fabric does not support datetimeoffset (for timezones)
        # So we will convert the timestamp to UTC and then convert it to a string
        return f"'{datetime.astimezone(timezone.utc).isoformat()}'"

    def _build_create_table_column_type(self, create_table_column: CREATE_TABLE_COLUMN) -> str:
        assert isinstance(create_table_column.type, SqlDataType)
        if create_table_column.type.name == "datetime2" and create_table_column.type.datetime_precision is None:
            return "datetime2(6)"
        elif create_table_column.type.name == "time" and create_table_column.type.datetime_precision is None:
            return "time(6)"
        else:
            return create_table_column.type.get_sql_data_type_str_with_parameters()

    def get_sql_data_type_name_by_soda_data_type_names(self) -> dict[str, str]:
        types = super().get_sql_data_type_name_by_soda_data_type_names()

        types[SodaDataTypeName.TEXT] = "varchar"
        # Fabric does not support datetimeoffset and does not have a timezone aware datetime type
        types[SodaDataTypeName.TIMESTAMP_TZ] = "datetime2"

        return types

    def get_data_source_data_type_name_by_soda_data_type_names(self) -> dict:
        result = super().get_data_source_data_type_name_by_soda_data_type_names()
        result[SodaDataTypeName.TIMESTAMP_TZ] = "datetime2"
        return result
